##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

from urllib import urlencode

import zope.interface
import zope.i18nmessageid
from zope.traversing.browser import absoluteURL

from z3c.table import interfaces
from z3c.batching.batch import first_neighbours_last

_ = zope.i18nmessageid.MessageFactory('z3c')


class BatchProvider(object):
    """Batch content provider.

    A batch provider is responsible for rendering the batch HTML and not for
    batching. The batch setup is directly done in the table. A batch provider
    get only used if the table rows is a batch.

    This batch provider offers a batch presentation for a given table. The
    batch provides different configuration options which can be overriden in
    custom implementations:
    
    The batch acts like this. If we have more batches than
    (prevBatchSize + nextBatchSize + 3) then the advanced batch subset is used.
    Otherwise, we will render all batch links.    
    Note, the additional factor 3 is the placeholder for the first, current and
    last item.

    Such a batch looks like:

    Renders the link for the first batch, spacers, the amount of links for 
    previous batches, the current batch link, spacers, the amount of links for 
    previous batches and the link for the last batch.
    
    Sample for 1000 items with 100 batches with batchSize of 10 and a     
    prevBatchSize of 3 and a nextBatchSize of 3:

    For the first item:
    [*1*][2][3][4] ... [100]

    In the middle:
    [1] ... [6][7][8][*9*][10][11][12] ... [100]

    At the end:
    [1] ... [97][98][99][*100*]
    """

    zope.interface.implements(interfaces.IBatchProvider)

    batchItems = []

    prevBatchSize = 3
    nextBatchSize = 3
    batchSpacer = u'...'

    _request_args = ['%(prefix)s-sortOn', '%(prefix)s-sortOrder']

    def __init__(self, context, request, table):
        self.__parent__ = context
        self.context = context
        self.request = request
        self.table = table
        self.batch = table.rows
        self.batches = table.rows.batches

    def getQueryStringArgs(self):
        """Collect additional terms from the request to include in links.

        API borrowed from z3c.table.header.ColumnHeader.
        """
        args = {}
        for key in self._request_args:
            key = key % dict(prefix=self.table.prefix)
            value = self.request.get(key, None)
            if value:
                args.update({key: value})
        return args

    def renderBatchLink(self, batch, cssClass=None):
        args = self.getQueryStringArgs()
        args[self.table.prefix +'-batchStart'] = batch.start
        args[self.table.prefix +'-batchSize'] = batch.size
        query = urlencode(sorted(args.items()))
        tableURL = absoluteURL(self.table, self.request)
        idx = batch.index + 1
        css = ' class="%s"' % cssClass
        cssClass = cssClass and css or u''
        return '<a href="%s?%s"%s>%s</a>' % (tableURL, query, cssClass, idx)

    def update(self):
        # 3 is is the placeholder for the first, current and last item.
        total = self.prevBatchSize + self.nextBatchSize + 3
        if self.batch.total <= total:
            # give all batches
            self.batchItems = self.batch.batches
        else:
            # switch to an advanced batch subset
            self.batchItems = first_neighbours_last(self.batches,
                                                    self.batch.index,
                                                    self.prevBatchSize,
                                                    self.nextBatchSize,
                                                    )

    def render(self):
        self.update()
        res = []
        append = res.append
        idx = 0
        lastIdx = len(self.batchItems)
        for batch in self.batchItems:
            idx += 1
            # build css class
            cssClasses = []
            if batch and batch == self.batch:
                cssClasses.append('current')
            if idx == 1:
                cssClasses.append('first')
            if idx == lastIdx:
                cssClasses.append('last')
            
            if cssClasses:
                css = ' '.join(cssClasses)
            else:
                css = None

            # render spaces
            if batch is None:
                append(self.batchSpacer)
            elif idx == 1:
                # render first
                append(self.renderBatchLink(batch, css))
            elif batch == self.batch:
                # render current
                append(self.renderBatchLink(batch, css))
            elif idx == lastIdx:
                # render last
                append(self.renderBatchLink(batch, css))
            else:
                append(self.renderBatchLink(batch))
        return u'\n'.join(res)



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from z3c.table.i18n import _
import zope.contentprovider.interfaces
import zope.interface
import zope.schema


class IValues(zope.interface.Interface):
    """Table value adapter."""

    values = zope.interface.Attribute('Iterable table row data sequence.')


class ITable(zope.contentprovider.interfaces.IContentProvider):
    """Table provider"""

    columnCounter = zope.schema.Int(
        title=_(u'Column counter'),
        description=_(u'Column counter'),
        default=0)

    columnIndexById = zope.interface.Attribute(
        'Dict of column index number by id')

    columnByName = zope.interface.Attribute('Dict of columns by name')

    columns = zope.interface.Attribute('Sequence of columns')

    rows = zope.interface.Attribute('Sequence of rows')

    selectedItems = zope.interface.Attribute('Sequence of selected items')

    # customize this part if needed
    prefix = zope.schema.BytesLine(
        title=_('Prefix'),
        description=_('The prefix of the table used to uniquely identify it.'),
        default='table')

    # css classes
    cssClasses = zope.interface.Attribute(
        'Dict of element name and CSS classes')

    # additional (row) css
    cssClassEven = zope.schema.TextLine(
        title=u'Even css row class',
        description=(u'CSS class for even rows.'),
        default=u'even',
        required=False)

    cssClassOdd = zope.schema.TextLine(
        title=u'Odd css row class',
        description=(u'CSS class for odd rows.'),
        default=u'odd',
        required=False)

    cssClassSelected = zope.schema.TextLine(
        title=u'Selected css row class',
        description=(u'CSS class for selected rows.'),
        default=u'selected',
        required=False)

    # sort attributes
    sortOn = zope.schema.Int(
        title=_(u'Sort on table index'),
        description=_(u'Sort on table index'),
        default=0)

    sortOrder = zope.schema.TextLine(
        title=_(u'Sort order'),
        description=_(u'Row sort order'),
        default=u'ascending')

    reverseSortOrderNames = zope.schema.List(
        title=u'Selected css row class',
        description=(u'CSS class for selected rows.'),
        value_type=zope.schema.TextLine(
            title=_(u'Reverse sort order name'),
            description=_(u'Reverse sort order name')
        ),
        default=[u'descending', u'reverse', u'down'],
        required=False)

    # batch attributes
    batchStart = zope.schema.Int(
        title=_(u'Batch start index'),
        description=_(u'Index the batch starts with'),
        default=0)

    batchSize = zope.schema.Int(
        title=_(u'Batch size'),
        description=_(u'The batch size'),
        default=50)

    startBatchingAt = zope.schema.Int(
        title=_(u'Batch start size'),
        description=_(u'The minimal size the batch starts to get used'),
        default=50)

    values = zope.interface.Attribute('Iterable table row data sequence.')

    def getCSSClass(element, cssClass=None):
        """Returns the css class if any or an empty string."""

    def setUpColumns():
        """Setup table column renderer."""

    def updateColumns():
        """Update columns."""

    def initColumns():
        """Initialize columns definitions used by the table"""

    def orderColumns():
        """Order columns."""

    def setUpRow(item):
        """Setup table row."""

    def setUpRows():
        """Setup table rows."""

    def getSortOn():
        """Returns sort on column id."""

    def getSortOrder():
        """Returns sort order criteria."""

    def sortRows():
        """Sort rows."""

    def getBatchSize():
        """Returns the batch size."""

    def getBatchStart():
        """Returns the batch start index."""

    def batchRows():
        """Batch rows."""

    def isSelectedRow(row):
        """Returns True for selected row."""

    def renderTable():
        """Render the table."""

    def renderHead():
        """Render the thead."""

    def renderHeadRow():
        """Render the table header rows."""

    def renderHeadCell(column):
        """Setup the table header rows."""

    def renderBody():
        """Render the table body."""

    def renderRows():
        """Render the table body rows."""

    def renderRow(row, cssClass=None):
        """Render the table body rows."""

    def renderCell(item, column, colspan=0):
        """Render a single table body cell."""

    def render():
        """Plain render method without keyword arguments."""


class ISequenceTable(ITable):
    """Sequence table adapts a sequence as context.

    This table can be used for adapting a z3c.indexer.search.ResultSet or
    z3c.batching.batch.Batch instance as context. Batch which wraps a
    ResultSet sequence.
    """


class IColumn(zope.interface.Interface):
    """Column provider"""

    id = zope.schema.TextLine(
        title = _(u'Id'),
        description = _(u'The column id'),
        default = None)

    # customize this part if needed
    colspan = zope.schema.Int(
        title = _(u'Colspan'),
        description = _(u'The colspan value'),
        default = 0)

    weight = zope.schema.Int(
        title = _(u'Weight'),
        description = _(u'The column weight'),
        default = 0)

    header = zope.schema.TextLine(
        title = _(u'Header name'),
        description = _(u'The header name'),
        default = u'')

    cssClasses = zope.interface.Attribute('Dict of element name and CSS classes')

    def getColspan(item):
        """Colspan value based on the given item."""

    def renderHeadCell():
        """Render the column header label."""

    def renderCell(item):
        """Render the column content."""


class INoneCell(IColumn):
    """None cell used for colspan."""


class IBatchProvider(zope.contentprovider.interfaces.IContentProvider):
    """Batch content provider"""

    def renderBatchLink(batch, cssClass=None):
        """Renders batch links."""

    def render():
        """Plain render method without keyword arguments."""


class IColumnHeader(zope.interface.Interface):
    """Multi-adapter for header rendering."""

    def update():
        """Override this method in subclasses if required"""

    def render():
        """Override this method in subclasses"""

    def getQueryStringArgs():
        """
        Because the header will most often be used to add links for sorting the
        columns it may also be necessary to collect other query arguments from
        the request.

        The initial use case here is to maintain a search term.
        """



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import unittest
import doctest
from zope.publisher.browser import TestRequest

import z3c.testing
from z3c.batching.batch import Batch
from z3c.table import testing
from z3c.table import interfaces
from z3c.table import table
from z3c.table import column
from z3c.table import batch


class FakeContainer(object):

    def values(self):
        pass


# table
class TestTable(z3c.testing.InterfaceBaseTest):

    def setUp(test):
        testing.setUpAdapters()

    def getTestInterface(self):
        return interfaces.ITable

    def getTestClass(self):
        return table.Table

    def getTestPos(self):
        return (FakeContainer(), TestRequest())


class TestSequenceTable(z3c.testing.InterfaceBaseTest):

    def setUp(test):
        testing.setUpAdapters()

    def getTestInterface(self):
        return interfaces.ITable

    def getTestClass(self):
        return table.SequenceTable

    def getTestPos(self):
        return (None, TestRequest())


# column
class TestColumn(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.IColumn

    def getTestClass(self):
        return column.Column

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        return ({}, TestRequest(), t)


class TestNoneCell(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.INoneCell

    def getTestClass(self):
        return column.NoneCell

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        return ({}, TestRequest(), t)


class TestNameColumn(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.IColumn

    def getTestClass(self):
        return column.NameColumn

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        return ({}, TestRequest(), t)


class TestRadioColumn(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.IColumn

    def getTestClass(self):
        return column.RadioColumn

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        return ({}, TestRequest(), t)


class TestCheckBoxColumn(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.IColumn

    def getTestClass(self):
        return column.CheckBoxColumn

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        return ({}, TestRequest(), t)


# batch
class TestBatchProvider(z3c.testing.InterfaceBaseTest):

    def getTestInterface(self):
        return interfaces.IBatchProvider

    def getTestClass(self):
        return batch.BatchProvider

    def getTestPos(self):
        t = table.Table(None, TestRequest())
        t.rows = Batch([])
        return ({}, TestRequest(), t)


def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite('README.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite('sequence.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite('sort.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite('batch.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS|doctest.REPORT_UDIFF,
            ),
        doctest.DocFileSuite('miscellaneous.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite('column.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        unittest.makeSuite(TestTable),
        unittest.makeSuite(TestSequenceTable),
        unittest.makeSuite(TestColumn),
        unittest.makeSuite(TestNoneCell),
        unittest.makeSuite(TestNameColumn),
        unittest.makeSuite(TestRadioColumn),
        unittest.makeSuite(TestCheckBoxColumn),
        unittest.makeSuite(TestBatchProvider),
        ))


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

from xml.sax.saxutils import quoteattr

import zope.interface
import zope.component
import zope.location

from z3c.batching.interfaces import IBatch
from z3c.batching.batch import Batch
from z3c.table import interfaces
from z3c.table import column


def getWeight(column):
    try:
        return int(column.weight)
    except AttributeError:
        return 0


def getSortMethod(idx):

    def getSortKey(item):
        sublist = item[idx]

        def getColumnSortKey(sublist):
            return sublist[1].getSortKey(sublist[0])

        return getColumnSortKey(sublist)

    return getSortKey


def nameColumn(column, name):
    """Give a column a __name__."""
    column.__name__ = name
    return column


class Table(zope.location.Location):
    """Generic usable table implementation."""

    zope.interface.implements(interfaces.ITable)

    # customize this part if needed
    prefix = 'table'

    # css classes
    cssClasses = {}
    # additional (row) css
    cssClassEven = u''
    cssClassOdd = u''
    cssClassSelected = u''

    # sort attributes
    sortOn = 0
    sortOrder = u'ascending'
    reverseSortOrderNames = [u'descending', u'reverse', u'down']

    # batch attributes
    batchProviderName = 'batch'
    batchStart = 0
    batchSize = 50
    startBatchingAt = 50

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.__parent__ = context
        # private variables will be set in update call
        self.batchProvider = None
        self.columnCounter = 0
        self.columnIndexById = {}
        self.columnByName = {}
        self.columns = None
        self.rows = []
        self.selectedItems = []

    def initColumns(self):
        # setup columns
        self.columns = self.setUpColumns()
        # order columns
        self.orderColumns()

    def getCSSClass(self, element, cssClass=None):
        klass = self.cssClasses.get(element)
        if klass and cssClass:
            klass = '%s %s' % (cssClass, klass)
        elif cssClass:
            klass = cssClass
        return klass and ' class=%s' % quoteattr(klass) or ''

    @property
    def values(self):
        adapter = zope.component.getMultiAdapter(
            (self.context, self.request, self), interfaces.IValues)
        return adapter.values

# setup

    def setUpColumns(self):
        cols = list(zope.component.getAdapters(
            (self.context, self.request, self), interfaces.IColumn))
        # use the adapter name as column name
        return [nameColumn(col, name) for name, col in cols]

    def updateColumns(self):
        for col in self.columns:
            col.update()

    def orderColumns(self):
        self.columnCounter = 0
        self.columns = sorted(self.columns, key=getWeight)
        for col in self.columns:
            self.columnByName[col.__name__] = col
            idx = self.columnCounter
            col.id = '%s-%s-%s' % (self.prefix, col.__name__, idx)
            self.columnIndexById[col.id] = idx
            self.columnCounter += 1

    def setUpRow(self, item):
        cols = []
        append = cols.append
        colspanCounter = 0
        countdown = len(self.columns)
        for col in self.columns:
            countdown -= 1
            colspan = 0
            if colspanCounter == 0:
                colspan = colspanCounter = col.getColspan(item)
                # adjust colspan because we define 0, 2, 3, etc.
                if colspanCounter > 0:
                    colspanCounter -=1

            if colspan == 0 and colspanCounter > 0:
                # override col if colspan is 0 and colspan coutner not 0
                colspanCounter -=1
                colspan = 0
                # now we are ready to setup dummy colspan cells
                col = column.NoneCell(self.context, self.request, self)

            # we reached the end of the table and have still colspan
            if (countdown - colspan) < 0:
                raise ValueError(
                    "Colspan for column '%s' is larger than the table." % col)

            append((item, col, colspan))
        return cols

    def setUpRows(self):
        return [self.setUpRow(item) for item in self.values]

# sort

    def getSortOn(self):
        """Returns sort on column id."""
        return self.request.get(self.prefix +'-sortOn', self.sortOn)

    def getSortOrder(self):
        """Returns sort order criteria."""
        return self.request.get(self.prefix +'-sortOrder',
            self.sortOrder)

    def sortRows(self):
        if self.sortOn is not None and self.rows and self.columns:
            sortOnIdx = self.columnIndexById.get(self.sortOn, 0)
            sortKeyGetter = getSortMethod(sortOnIdx)
            rows = sorted(self.rows, key=sortKeyGetter)
            if self.sortOrder in self.reverseSortOrderNames:
                rows.reverse()
            self.rows = rows

# batch

    def getBatchSize(self):
        return int(self.request.get(self.prefix +'-batchSize',
            self.batchSize))

    def getBatchStart(self):
        return int(self.request.get(self.prefix +'-batchStart',
            self.batchStart))

    def batchRows(self):
        if len(self.rows) > self.startBatchingAt:
            self.rows = Batch(self.rows, start=self.batchStart,
                size=self.batchSize)

    def updateBatch(self):
        if IBatch.providedBy(self.rows):
            self.batchProvider = zope.component.getMultiAdapter((self.context,
                self.request, self), interfaces.IBatchProvider,
                name=self.batchProviderName)
            self.batchProvider.update()

    def isSelectedRow(self, row):
        item, col, colspan = row[0]
        if item in self.selectedItems:
            return True
        return False

# render

    def renderBatch(self):
        if self.batchProvider is None:
            return u''
        return self.batchProvider.render()

    def renderTable(self):
        if self.columns:
            cssClass = self.getCSSClass('table')
            head = self.renderHead()
            body = self.renderBody()
            return '<table%s>%s%s\n</table>' % (cssClass, head, body)
        return u''

    def renderHead(self):
        cssClass = self.getCSSClass('thead')
        rStr = self.renderHeadRow()
        return '\n  <thead%s>%s\n  </thead>' % (cssClass, rStr)

    def renderHeadRow(self):
        cssClass = self.getCSSClass('tr')
        cells = [self.renderHeadCell(col) for col in self.columns]
        return u'\n    <tr%s>%s\n    </tr>' % (cssClass, u''.join(cells))

    def renderHeadCell(self, column):
        cssClass = column.cssClasses.get('th')
        cssClass = self.getCSSClass('th', cssClass)
        return u'\n      <th%s>%s</th>' % (cssClass, column.renderHeadCell())

    def renderBody(self):
        cssClass = self.getCSSClass('tbody')
        rStr = self.renderRows()
        return '\n  <tbody%s>%s\n  </tbody>' % (cssClass, rStr)

    def renderRows(self):
        counter = 0
        rows = []
        cssClasses = (self.cssClassEven, self.cssClassOdd)
        append = rows.append
        for row in self.rows:
            append(self.renderRow(row, cssClasses[counter % 2]))
            counter +=1
        return u''.join(rows)

    def renderRow(self, row, cssClass=None):
        isSelected = self.isSelectedRow(row)
        if isSelected and self.cssClassSelected and cssClass:
            cssClass = '%s %s' % (self.cssClassSelected, cssClass)
        elif isSelected and self.cssClassSelected:
            cssClass = self.cssClassSelected
        cssClass = self.getCSSClass('tr', cssClass)
        cells = [self.renderCell(item, col, colspan)
                 for item, col, colspan in row]
        return u'\n    <tr%s>%s\n    </tr>' % (cssClass, u''.join(cells))

    def renderCell(self, item, column, colspan=0):
        if interfaces.INoneCell.providedBy(column):
            return u''
        cssClass = column.cssClasses.get('td')
        cssClass = self.getCSSClass('td', cssClass)
        colspanStr = colspan and ' colspan="%s"' % colspan or ''
        return u'\n      <td%s%s>%s</td>' % (cssClass, colspanStr,
            column.renderCell(item))

    def update(self):
        # reset values
        self.columnCounter = 0
        self.columnByIndex = {}
        self.selectedItems = []

        # use batch values from request or the existing ones
        self.batchSize = self.getBatchSize()
        self.batchStart = self.getBatchStart()
        # use sorting values from request or the existing ones
        self.sortOn = self.getSortOn()
        self.sortOrder = self.getSortOrder()

        # initialize columns
        self.initColumns()

        # update columns
        self.updateColumns()

        # setup headers based on columns
        self.rows = self.setUpRows()

        # sort items on columns
        self.sortRows()

        # batch sorted rows
        self.batchRows()

        self.updateBatch()

    def render(self):

        # allow to use a template for rendering the table, this will allow
        # to position the batch before and after the table

        return self.renderTable()

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


class SequenceTable(Table):
    """Sequence table adapts a sequence as context.

    This table can be used for adapting a z3c.indexer.search.ResultSet or
    z3c.batching.batch.Batch instance as context. Batch which wraps a
    ResultSet sequence.
    """

    zope.interface.implements(interfaces.ISequenceTable)



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import zope.interface
from zope.publisher.interfaces.browser import IBrowserRequest

from z3c.table import interfaces


class ValuesMixin(object):
    """Mixin for different value adapters."""

    zope.interface.implements(interfaces.IValues)

    def __init__(self, context, request, table):
        self.context = context
        self.request = request
        self.table = table


class ValuesForContainer(ValuesMixin):
    """Values from a simple IContainer."""

    zope.component.adapts(zope.interface.Interface, IBrowserRequest,
        interfaces.ITable)

    @property
    def values(self):
        return self.context.values()


class ValuesForSequence(ValuesMixin):
    """Values from a simple IContainer."""

    zope.component.adapts(zope.interface.Interface, IBrowserRequest,
        interfaces.ISequenceTable)


    @property
    def values(self):
        return self.context


##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

from urllib import urlencode
from z3c.table import interfaces
from zope.dublincore.interfaces import IZopeDublinCore
from zope.security.interfaces import Unauthorized
from zope.traversing import api
from zope.traversing.browser import absoluteURL
import zope.i18n
import zope.i18nmessageid
import zope.interface
import zope.location

_ = zope.i18nmessageid.MessageFactory('z3c')


def addColumn(self, class_, name, cellRenderer=None, headCellRenderer=None,
    colspan= None, weight=None, header=None, cssClasses=None, **kws):
    if not interfaces.IColumn.implementedBy(class_):
        raise ValueError('class_ %s must implement IColumn.' % class_)
    column = class_(self.context, self.request, self)
    column.__parent__ = self
    column.__name__ = name
    if cellRenderer is not None:
        # overload method
        column.renderCell = cellRenderer
    if headCellRenderer is not None:
        # overload method
        column.renderHeadCell = headCellRenderer
    if colspan is not None:
        column.colspan = colspan
    if weight is not None:
        column.weight = weight
    if header is not None:
        column.header = header
    if cssClasses is not None:
        column.cssClasses = cssClasses
    for name, value in kws.items():
        setattr(column, name, value)
    return column


def getName(item):
    # probably not IPhysicallyLocatable but still could have a __name__
    try:
        return api.getName(item)
    except TypeError, e:
        return item.__name__


def safeGetAttr(obj, attr, default):
    try:
        return getattr(obj, attr, default)
    except Unauthorized:
        return default


class Column(zope.location.Location):
    """Column provider."""

    zope.interface.implements(interfaces.IColumn)

    # variables will be set by table
    id = None

    # customize this part if needed
    colspan = 0
    weight = 0
    header = u''
    cssClasses = {}

    def __init__(self, context, request, table):
        self.__parent__ = context
        self.context = context
        self.request = request
        self.table = table

    def update(self):
        pass

    def getColspan(self, item):
        """Returns the colspan value."""
        return self.colspan

    def getSortKey(self, item):
        """Returns the sort key used for column sorting."""
        return self.renderCell(item)

    def renderHeadCell(self):
        """Header cell content."""
        header = zope.component.queryMultiAdapter((self.context,
            self.request, self.table, self), interfaces.IColumnHeader)
        if header:
            header.update()
            return header.render()
        return zope.i18n.translate(self.header, context=self.request)

    def renderCell(self, item):
        """Cell content."""
        raise NotImplementedError('Subclass must implement renderCell')

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


class NoneCell(Column):
    """None cell is used for mark a previous colspan."""

    zope.interface.implements(interfaces.INoneCell)

    def getColspan(self, item):
        return 0

    def renderHeadCell(self):
        return u''

    def renderCell(self, item):
        return u''


# predefined columns
class NameColumn(Column):
    """Name column."""

    header = _('Name')

    def renderCell(self, item):
        return getName(item)


class RadioColumn(Column):
    """Radio column."""

    header = _('X')

    @apply
    def selectedItem():
        # use the items form the table
        def get(self):
            if len(self.table.selectedItems):
                return list(self.table.selectedItems).pop()
        def set(self, value):
            self.table.selectedItems = [value]
        return property(get, set)

    def getSortKey(self, item):
        return getName(item)

    def getItemKey(self, item):
        return '%s-selectedItem' % self.id

    def getItemValue(self, item):
        return getName(item)

    def update(self):
        items = [item for item in self.table.values
                 if self.getItemValue(item) in self.request.get(
                     self.getItemKey(item), [])]
        if len(items):
            self.selectedItem = items.pop()

    def renderCell(self, item):
        selected = u''
        if item == self.selectedItem:
            selected='checked="checked"'
        return u'<input type="radio" class="%s" name="%s" value="%s" %s />' %(
            'radio-widget', self.getItemKey(item), self.getItemValue(item),
            selected)


class CheckBoxColumn(Column):
    """Checkbox column."""

    header = _('X')
    weight = 10

    @apply
    def selectedItems():
        # use the items form the table
        def get(self):
            return self.table.selectedItems
        def set(self, values):
            self.table.selectedItems = values
        return property(get, set)

    def getSortKey(self, item):
        return getName(item)

    def getItemKey(self, item):
        return '%s-selectedItems' % self.id

    def getItemValue(self, item):
        return getName(item)

    def isSelected(self, item):
        v = self.request.get(self.getItemKey(item), [])
        if not isinstance(v, list):
            # ensure that we have a list which prevents to compare strings
            v = [v]
        if self.getItemValue(item) in v:
            return True
        return False

    def update(self):
        self.selectedItems = [item for item in self.table.values
                              if self.isSelected(item)]

    def renderCell(self, item):
        selected = u''
        if item in self.selectedItems:
            selected='checked="checked"'
        return u'<input type="checkbox" class="%s" name="%s" value="%s" %s />' \
            %('checkbox-widget', self.getItemKey(item), self.getItemValue(item),
            selected)


class GetAttrColumn(Column):
    """Get attribute column."""

    attrName = None
    defaultValue = u''

    def getValue(self, obj):
        if obj is not None and self.attrName is not None:
            return safeGetAttr(obj, self.attrName, self.defaultValue)
        return self.defaultValue

    def renderCell(self, item):
        return self.getValue(item)


class I18nGetAttrColumn(GetAttrColumn):
    """GetAttrColumn which translates its content."""

    def renderCell(self, item):
        return zope.i18n.translate(self.getValue(item), context=self.request)


class FormatterColumn(Column):
    """Formatter column."""

    formatterCategory = u'dateTime'
    formatterLength = u'medium'
    formatterName = None
    formatterCalendar = u'gregorian'

    def getFormatter(self):
        return self.request.locale.dates.getFormatter(
            self.formatterCategory, self.formatterLength, self.formatterName,
            self.formatterCalendar)


class GetAttrFormatterColumn(FormatterColumn, GetAttrColumn):
    """Get attribute and formatter column."""

    def renderCell(self, item):
        formatter = self.getFormatter()
        value = self.getValue(item)
        if value:
            value = formatter.format(value)
        return value


class CreatedColumn(FormatterColumn, GetAttrColumn):
    """Created date column."""

    header = _('Created')
    weight = 100

    formatterCategory = u'dateTime'
    formatterLength = u'short'
    attrName = 'created'

    def renderCell(self, item):
        formatter = self.getFormatter()
        dc = IZopeDublinCore(item, None)
        value = self.getValue(dc)
        if value:
            value = formatter.format(value)
        return value


class ModifiedColumn(FormatterColumn, GetAttrColumn):
    """Created date column."""

    header = _('Modified')
    weight = 110

    formatterCategory = u'dateTime'
    formatterLength = u'short'
    attrName = 'modified'

    def renderCell(self, item):
        formatter = self.getFormatter()
        dc = IZopeDublinCore(item, None)
        value = self.getValue(dc)
        if value:
            value = formatter.format(value)
        return value


class LinkColumn(Column):
    """Name column."""

    header = _('Name')
    linkName = None
    linkTarget = None
    linkContent = None
    linkCSS = None

    def getLinkURL(self, item):
        """Setup link url."""
        if self.linkName is not None:
            return '%s/%s' % (absoluteURL(item, self.request), self.linkName)
        return absoluteURL(item, self.request)

    def getLinkCSS(self, item):
        """Setup link css."""
        return self.linkCSS and ' class="%s"' % self.linkCSS or ''

    def getLinkTarget(self, item):
        """Setup link css."""
        return self.linkTarget and ' target="%s"' % self.linkTarget or ''

    def getLinkContent(self, item):
        """Setup link content."""
        if self.linkContent:
            return zope.i18n.translate(self.linkContent, context=self.request)
        return getName(item)

    def renderCell(self, item):
        # setup a tag
        return '<a href="%s"%s%s>%s</a>' % (self.getLinkURL(item),
            self.getLinkTarget(item), self.getLinkCSS(item),
            self.getLinkContent(item))


class EMailColumn(LinkColumn, GetAttrColumn):
    "Column to display mailto links."

    header = _(u'E-Mail')
    attrName = None # attribute name which contains the mail address
    defaultValue = u'' # value which is rendered when no value is found
    linkContent = None

    def getLinkURL(self, item):
        return 'mailto:%s' % self.getValue(item)

    def getLinkContent(self, item):
        if self.linkContent:
            return zope.i18n.translate(self.linkContent, context=self.request)
        return self.getValue(item)

    def renderCell(self, item):
        value = self.getValue(item)
        if value is self.defaultValue or value is None:
            return self.defaultValue
        return super(EMailColumn, self).renderCell(item)

def ensureList(item):
    if not isinstance(item, (list, tuple)):
        return [item]
    return item

class SelectedItemColumn(LinkColumn):
    """Link which can set an item."""

    selectedItem = None

    @property
    def viewURL(self):
        return '%s/%s' % (absoluteURL(self.context, self.request),
            self.table.__name__)

    def getItemKey(self, item):
        return '%s-selectedItems' % self.id

    def getItemValue(self, item):
        return getName(item)

    def getSortKey(self, item):
        """Returns the sort key used for column sorting."""
        return self.getLinkContent(item)

    def getLinkContent(self, item):
        """Setup link content."""
        return self.linkContent or getName(item)

    def getLinkURL(self, item):
        """Setup link url."""
        return '%s?%s' % (self.viewURL,
            urlencode({self.getItemKey(item): self.getItemValue(item)}))

    def update(self):
        items = [item for item in self.table.values
                 if self.getItemValue(item) in ensureList(self.request.get(
                     self.getItemKey(item), []))]
        if len(items):
            self.selectedItem = items.pop()
            self.table.selectedItems = [self.selectedItem]


class ContentsLinkColumn(LinkColumn):
    """Link pointing to contents.html."""

    linkName = 'contents.html'


class IndexLinkColumn(LinkColumn):
    """Link pointing to index.html."""

    linkName = 'index.html'


class EditLinkColumn(LinkColumn):
    """Link pointing to edit.html."""

    linkName = 'edit.html'



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: header.py 115256 2010-07-31 12:29:38Z icemac $
"""
__docformat__ = "reStructuredText"


from z3c.table.i18n import _
import urllib
import z3c.table.interfaces
import zope.i18n
import zope.interface


class ColumnHeader(object):
    """ColumnHeader renderer provider"""

    zope.interface.implements(z3c.table.interfaces.IColumnHeader)

    _request_args = []

    def __init__(self, context, request, table, column):
        self.__parent__ = context
        self.context = context
        self.request = request
        self.table = table
        self.column = column

    def update(self):
        """Override this method in subclasses if required"""
        pass

    def render(self):
        """Override this method in subclasses"""
        return self.column.header

    def getQueryStringArgs(self):
        """
        Collect additional terms from the request and include in sorting column
        headers

        Perhaps this should be in separate interface only for sorting headers?

        """
        args = {}
        for key in self._request_args:
            value = self.request.get(key, None)
            if value:
                args.update({key: value})
        return args


class SortingColumnHeader(ColumnHeader):
    """Sorting column header."""

    def render(self):
        table = self.table
        prefix = table.prefix
        colID = self.column.id

        # this may return a string 'id-name-idx' if coming from request,
        # otherwise in Table class it is intialised as a integer string
        currentSortID = table.getSortOn()
        try:
            currentSortID = int(currentSortID)
        except ValueError:
            currentSortID = currentSortID.rsplit('-', 1)[-1]

        currentSortOrder = table.getSortOrder()

        sortID = colID.rsplit('-', 1)[-1]

        sortOrder = table.sortOrder
        if int(sortID) == int(currentSortID):
            # ordering the same column so we want to reverse the order
            if currentSortOrder in table.reverseSortOrderNames:
                sortOrder = 'ascending'
            elif currentSortOrder == 'ascending':
                sortOrder = table.reverseSortOrderNames[0]

        args = self.getQueryStringArgs()
        args.update({'%s-sortOn' % prefix: colID,
                     '%s-sortOrder' % prefix: sortOrder})
        queryString = '?%s' % (urllib.urlencode(args))

        return '<a href="%s" title="%s">%s</a>' % (
            queryString,
            zope.i18n.translate(_('Sort'), context=self.request),
            zope.i18n.translate(self.column.header, context=self.request))




##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import datetime
import zope.interface
import zope.component
from zope.container import btree
from zope.dublincore.interfaces import IZopeDublinCore
from zope.security import checker
from zope.app.testing import setup

from z3c.table import column, table
import z3c.table.value


class TitleColumn(column.Column):

    weight = 10
    header = u'Title'

    def renderCell(self, item):
        return u'Title: %s' % item.title


class NumberColumn(column.Column):

    header = u'Number'
    weight = 20

    def getSortKey(self, item):
        return item.number

    def renderCell(self, item):
        return 'number: %s' % item.number


class Container(btree.BTreeContainer):
    """Sample container."""
    __name__ = u'container'


class Content(object):
    """Sample content."""

    def __init__(self, title, number):
        self.title = title
        self.number = number


class SimpleTable(table.Table):

    def setUpColumns(self):
        return [
            column.addColumn(self, TitleColumn, u'title',
                             cellRenderer=cellRenderer,
                             headCellRenderer=headCellRenderer,
                             weight=1),
            column.addColumn(self, NumberColumn, name=u'number',
                             weight=2, header=u'Number')]


def headCellRenderer():
    return u'My items'


def cellRenderer(item):
    return u'%s item' % item.title


class DublinCoreAdapterStub(object):
    """Dublin core adapter stub."""

    zope.interface.implements(IZopeDublinCore)
    zope.component.adapts(zope.interface.Interface)

    __Security_checker__ = checker.Checker(
        {"created": "zope.Public",
         "modified": "zope.Public",
         "title": "zope.Public",
         },
        {"title": "zope.app.dublincore.change"})

    def __init__(self, context):
        pass
    title = 'faux title'
    size = 1024
    created = datetime.datetime(2001, 1, 1, 1, 1, 1)
    modified = datetime.datetime(2002, 2, 2, 2, 2, 2)


def setUpAdapters():
    zope.component.provideAdapter(z3c.table.value.ValuesForContainer)
    zope.component.provideAdapter(z3c.table.value.ValuesForSequence)


def setUp(test):
    test.globs['root'] = setup.placefulSetUp(True)
    zope.component.provideAdapter(DublinCoreAdapterStub)
    setUpAdapters()


def tearDown(test):
    setup.placefulTearDown()
